OPENROUTER_BASE = "https://openrouter.ai"
OPENROUTER_API_BASE = f"{OPENROUTER_BASE}/api/v1"
OPENROUTER_REFERRER = "https://github.com/xvnpw/ai-threat-modeling-action"
OPENROUTER_API_KEY = "OPENROUTER_API_KEY"
OPENAI_API_KEY = "OPENAI_API_KEY"
ANTHROPIC_API_KEY = "ANTHROPIC_API_KEY"
GOOGLE_API_KEY = "GOOGLE_API_KEY"
DEFAULT_CONTEXT_WINDOW = 70000
DEFAULT_CHUNK_SIZE = 60000
